@@ -18,6 +18,10 @@ from typing import (
     Union,
 )
 
+import io
+import requests
+import gzip
+
 import geopandas as gpd
 import numpy as np
 import pandas as pd
@@ -36,6 +40,8 @@ from censusdis.impl.varcache import VariableCache
 from censusdis.impl.varsource.base import VintageType
 from censusdis.impl.varsource.censusapi import CensusApiVariableSource
 from censusdis.values import ALL_SPECIAL_VALUES
+from censusdis.datasets import ACS5, DECENNIAL_PUBLIC_LAW_94_171
+from censusdis.states import ABBREVIATIONS_FROM_IDS
 
 import censusdis.impl.fetch
 
@@ -113,10 +119,13 @@ def _download_multiple(
     vintage: VintageType,
     download_variables: List[str],
     *,
-    key: Optional[str],
+    query_filter: Dict[str, str],
+    api_key: Optional[str],
     census_variables: "VariableCache",
-    with_geometry: bool = False,
-    row_keys: Optional[Union[str, Iterable[str]]] = None,
+    with_geometry: bool,
+    with_geometry_columns: bool,
+    tiger_shapefiles_only: bool,
+    row_keys: Union[str, Iterable[str]],
     **kwargs: cgeo.InSpecType,
 ) -> pd.DataFrame:
     """
@@ -137,12 +146,35 @@ def _download_multiple(
         a timeseries data set, pass the string `'timeseries'`.
     download_variables
         The census variables to download, for example `["NAME", "B01001_001E"]`.
+    query_filter
+        A dictionary of values to filter on. For example, if
+        `query_filter={'NAICS2017': '72251'}` then only rows
+        where the variable `NAICS2017` has a value of `'72251'`
+        will be returned.
+
+        This filtering is done on the server side, not the client
+        side, so it is far more efficient than querying without a
+        query filter and then manually filtering the results.
     with_geometry
         If `True` a :py:class:`gpd.GeoDataFrame` will be returned and each row
         will have a geometry that is a cartographic boundary suitable for platting
         a map. See https://www.census.gov/geographies/mapping-files/time-series/geo/cartographic-boundary.2020.html
         for details of the shapefiles that will be downloaded on your behalf to
         generate these boundaries.
+    with_geometry_columns
+        If `True` keep all the additional columns that come with shapefiles
+        downloaded to get geometry information.
+    tiger_shapefiles_only
+        If `True` only look for TIGER shapefiles. If `False`, first look
+        for CB shapefiles
+        (https://www.census.gov/geographies/mapping-files/time-series/geo/carto-boundary-file.html),
+        which are more suitable for plotting maps, then fall back on the full
+        TIGER files
+        (https://www.census.gov/geographies/mapping-files/time-series/geo/tiger-line-file.html)
+        only if CB is not available. This is mainly set to `True` only
+        when `with_geometry_columns` is also set to `True`. The reason
+        is that the additional columns in the shapefiles are different
+        in the CB files than in the TIGER files.
     api_key
         An optional API key. If you don't have or don't use a key, the number
         of calls you can make will be limited.
@@ -195,15 +227,20 @@ def _download_multiple(
             "use download instead."
         )
 
-    # Get the data for each chunk.
+    # Get the data for each chunk. Note that we leave out
+    # extra geometry columns at this point. We will get them
+    # later if we need them, but they get in the way at this
+    # point.
     dfs = [
         download(
             dataset,
             vintage,
             variable_group,
-            api_key=key,
+            query_filter=query_filter,
+            api_key=api_key,
             variable_cache=census_variables,
             with_geometry=with_geometry and (ii == 0),
+            with_geometry_columns=False,
             **kwargs,
         )
         for ii, variable_group in enumerate(variable_groups)
@@ -216,6 +253,22 @@ def _download_multiple(
     # the census API.
     geo_key_variables = [f for f in dfs[0].columns if f not in set(variable_groups[0])]
 
+    # Now that we know the geometry keys, we may have to get back the other
+    # geometry columns we left out the first time.
+    if with_geometry and with_geometry_columns:
+        dfs[0] = download(
+            dataset,
+            vintage,
+            variable_groups[0],
+            query_filter=query_filter,
+            api_key=api_key,
+            variable_cache=census_variables,
+            with_geometry=with_geometry,
+            with_geometry_columns=with_geometry_columns,
+            tiger_shapefiles_only=tiger_shapefiles_only,
+            **kwargs,
+        )
+
     # If we put in the geometry column, it's not part of the
     # key.
     if with_geometry:
@@ -254,6 +307,17 @@ def _download_multiple(
             merge_strategy = False
             break
 
+    if with_geometry and with_geometry_columns and not merge_strategy:
+        raise ValueError(
+            "`with_geometry_columns=True` is only supported for very wide results "
+            "when the merge_strategy can be used. This merge strategy is used when every "
+            "row of the result is for a unique geography, as is the case in data sets like "
+            f'ACS5 ("{ACS5}") and DECENNIAL_PUBLIC_LAW_94_171 ("{DECENNIAL_PUBLIC_LAW_94_171}"). '
+            "If this functionality is really important to you (note that it would create a lot "
+            "of duplicate geoemetry values, we suggest you set `with_geometry=False` in this call "
+            "and then merge with a `GeoDatFrame` with the geometries you want after the fact."
+        )
+
     if merge_strategy:
         # We can do the merge strategy.
 
@@ -262,7 +326,17 @@ def _download_multiple(
         df_data = dfs[0]
 
         for df_right in dfs[1:]:
-            df_data = df_data.merge(df_right, on=merge_keys)
+            df_data_columns = set(df_data.columns)
+            df_data = df_data.merge(
+                df_right[
+                    [
+                        col
+                        for col in df_right.columns
+                        if (col in merge_keys) or (col not in df_data_columns)
+                    ]
+                ],
+                on=merge_keys,
+            )
     else:
         # We are going to have to fall back on the concat
         # strategy. Before we do the concat, however, let's
@@ -320,6 +394,198 @@ def _download_multiple(
     return df_data
 
 
+def download_lodes(
+    dataset: str,
+    vintage: VintageType,
+    download_variables: Optional[Union[str, Iterable[str]]] = None,
+    version: Optional[str] = None,
+    home_geography: Optional[Union[bool, Dict[str, str]]] = None,
+    with_geometry: bool = False,
+    with_geometry_columns: bool = False,
+    tiger_shapefiles_only: bool = False,
+    remove_water: bool = False,
+    **kwargs: cgeo.InSpecType,
+) -> Union[pd.DataFrame, gpd.GeoDataFrame]:
+    """
+    Download LODES data from the US Census API.
+
+    This is typically not called directly, but instead LODES data
+    is obtained by calling :py:func:`~download`, which then calls
+    this as needed for LODES data sets.
+
+    Parameters
+    ----------
+    dataset
+        The dataset to download from. For example `"acs/acs5"`,
+        `"dec/pl"`, or `"timeseries/poverty/saipe/schdist"`. There are
+        symbolic names for datasets, like `ACS5` for `"acs/acs5"
+        in :py:module:`censusdis.datasets`.
+    vintage
+        The vintage to download data for. For most data sets this is
+        an integer year, for example, `2020`. But for
+        a timeseries data set, pass the string `'timeseries'`.
+    download_variables
+        The census variables to download, for example `["NAME", "B01001_001E"]`.
+    with_geometry
+        If `True` a :py:class:`gpd.GeoDataFrame` will be returned and each row
+        will have a geometry that is a cartographic boundary suitable for platting
+        a map. See https://www.census.gov/geographies/mapping-files/time-series/geo/cartographic-boundary.2020.html
+        for details of the shapefiles that will be downloaded on your behalf to
+        generate these boundaries.
+    with_geometry_columns
+        If `True` keep all the additional columns that come with shapefiles
+        downloaded to get geometry information.
+    tiger_shapefiles_only
+        If `True` only look for TIGER shapefiles. If `False`, first look
+        for CB shapefiles
+        (https://www.census.gov/geographies/mapping-files/time-series/geo/carto-boundary-file.html),
+        which are more suitable for plotting maps, then fall back on the full
+        TIGER files
+        (https://www.census.gov/geographies/mapping-files/time-series/geo/tiger-line-file.html)
+        only if CB is not available. This is mainly set to `True` only
+        when `with_geometry_columns` is also set to `True`. The reason
+        is that the additional columns in the shapefiles are different
+        in the CB files than in the TIGER files.
+    remove_water
+        If `True` and if with_geometry=True, will query TIGER for AREAWATER shapefiles and
+        remove water areas from returned geometry.
+    """
+    if version is None:
+        version = "LODES8"
+
+    if isinstance(home_geography, bool):
+        if home_geography:
+            home_geography = dict(**kwargs)
+        else:
+            home_geography = None
+
+    bound_path = cgeo.PathSpec.partial_prefix_match(dataset, vintage, **kwargs)
+    geo_bindings = bound_path.bindings
+
+    state = geo_bindings["state"]
+
+    if state == "*":
+        # TODO - we could just concatenate them all.
+        raise ValueError("Wildcards not supported for state LODES data.")
+
+    if state not in ABBREVIATIONS_FROM_IDS:
+        raise ValueError(f"Unknown state id {state}")
+
+    state_name = ABBREVIATIONS_FROM_IDS[state].lower()
+
+    _, data_set_type, part_or_segment, job_type = dataset.split("/")
+
+    if data_set_type in ["rac", "wac"]:
+        part_or_segment = part_or_segment.upper()
+
+    url = (
+        f"https://lehd.ces.census.gov/data/lodes/{version}/{state_name}/{data_set_type}/{state_name}_{data_set_type}_"
+        f"{part_or_segment}_{job_type.upper()}_{vintage}.csv.gz"
+    )
+
+    logger.info(f"Downloading LODES data from {url}")
+
+    results = requests.get(url)
+    if results.status_code != requests.status_codes.codes.OK:
+        raise CensusApiException(
+            f"Unable to get LODES data. Attempted to fetch from {url}. "
+            f"Status: {results.status_code}; {results.reason}"
+        )
+
+    gz_content = requests.get(url).content
+    content = gzip.decompress(gz_content)
+    df_lodes = pd.read_csv(
+        io.StringIO(content.decode("utf-8")), dtype={"w_geocode": str, "h_geocode": str}
+    )
+
+    # We don't need the date.
+    df_lodes = df_lodes.drop("createdate", axis="columns")
+
+    # Map the geographies to the conventions censusdis uses.
+
+    def map_geo_cols(*, from_prefix: str, to_suffix: str = ""):
+        df_lodes[f"STATE{to_suffix}"] = df_lodes[f"{from_prefix}geocode"].str[:2]
+        df_lodes[f"COUNTY{to_suffix}"] = df_lodes[f"{from_prefix}geocode"].str[2:5]
+        df_lodes[f"TRACT{to_suffix}"] = df_lodes[f"{from_prefix}geocode"].str[5:11]
+        df_lodes[f"BLOCK{to_suffix}"] = df_lodes[f"{from_prefix}geocode"].str[11:15]
+
+    if data_set_type in ["od", "wac"]:
+        map_geo_cols(from_prefix="w_")
+    else:
+        map_geo_cols(from_prefix="h_")
+
+    if data_set_type == "od":
+        map_geo_cols(from_prefix="h_", to_suffix="_H")
+
+    for geocode_col in ["w_geocode", "h_geocode"]:
+        if geocode_col in df_lodes.columns:
+            df_lodes.drop(geocode_col, axis="columns")
+
+    group_keys = []
+    selectors = {}
+
+    for geo, binding in geo_bindings.items():
+        group_keys.append(f"{geo.upper()}")
+        if binding != "*":
+            selectors[f"{geo.upper()}"] = binding
+
+    if data_set_type == "od":
+        if home_geography is None:
+            for col in df_lodes.columns:
+                if col.endswith("_H"):
+                    group_keys.append(col)
+        else:
+            # There is more grouping to do.
+            home_bound_path = cgeo.PathSpec.partial_prefix_match(
+                dataset, vintage, **home_geography
+            )
+            home_geo_bindings = home_bound_path.bindings
+
+            for geo, binding in home_geo_bindings.items():
+                group_keys.append(f"{geo.upper()}_H")
+                if binding != "*":
+                    selectors[f"{geo.upper()}_H"] = binding
+
+    # Filter down based on fixed bindings.
+    if selectors:
+        criteria = None
+        for col, binding in selectors.items():
+            if criteria is None:
+                criteria = df_lodes[col] == binding
+            else:
+                criteria = criteria & (df_lodes[col] == binding)
+        df_lodes = df_lodes[criteria]
+
+    if download_variables is None:
+        download_variables = [
+            col for col in df_lodes.columns if df_lodes[col].dtype != object
+        ]
+
+    # Group based on group keys.
+    df_lodes = df_lodes.groupby(group_keys)[download_variables].sum().reset_index()
+
+    if with_geometry:
+        # We need to get the geometry and merge it in.
+        geo_level = bound_path.path_spec.path[-1]
+        shapefile_scope = bound_path.bindings[bound_path.path_spec.path[0]]
+
+        gdf_data = add_geography(
+            df_lodes,
+            vintage,
+            shapefile_scope,
+            geo_level,
+            with_geometry_columns=with_geometry_columns,
+            tiger_shapefiles_only=tiger_shapefiles_only,
+        )
+
+        if remove_water:
+            gdf_data = clip_water(gdf_data, vintage)
+
+        return gdf_data
+
+    return df_lodes
+
+
 def download(
     dataset: str,
     vintage: VintageType,
@@ -329,7 +595,10 @@ def download(
     leaves_of_group: Optional[Union[str, Iterable[str]]] = None,
     set_to_nan: Union[bool, Iterable[int]] = True,
     skip_annotations: bool = True,
+    query_filter: Optional[Dict[str, str]] = None,
     with_geometry: bool = False,
+    with_geometry_columns: bool = False,
+    tiger_shapefiles_only: bool = False,
     remove_water: bool = False,
     download_contained_within: Optional[Dict[str, cgeo.InSpecType]] = None,
     area_threshold: float = 0.8,
@@ -365,7 +634,7 @@ def download(
     dataset
         The dataset to download from. For example `"acs/acs5"`,
         `"dec/pl"`, or `"timeseries/poverty/saipe/schdist"`. There are
-        symbolic names for datasets, like `ACS5` for `"acs/acs5"
+        symbolic names for datasets, like `ACS5` for `"acs/acs5"`
         in :py:module:`censusdis.datasets`.
     vintage
         The vintage to download data for. For most data sets this is
@@ -393,12 +662,35 @@ def download(
         annotations rather than actual values. See :py:meth:`VariableCache.group_variables`
         for more details. Variable names passed in `download_variables` are not
         affected by this flag.
+    query_filter
+        A dictionary of values to filter on. For example, if
+        `query_filter={'NAICS2017': '72251'}` then only rows
+        where the variable `NAICS2017` has a value of `'72251'`
+        will be returned.
+
+        This filtering is done on the server side, not the client
+        side, so it is far more efficient than querying without a
+        query filter and then manually filtering the results.
     with_geometry
         If `True` a :py:class:`gpd.GeoDataFrame` will be returned and each row
         will have a geometry that is a cartographic boundary suitable for platting
         a map. See https://www.census.gov/geographies/mapping-files/time-series/geo/cartographic-boundary.2020.html
         for details of the shapefiles that will be downloaded on your behalf to
         generate these boundaries.
+    with_geometry_columns
+        If `True` keep all the additional columns that come with shapefiles
+        downloaded to get geometry information.
+    tiger_shapefiles_only
+        If `True` only look for TIGER shapefiles. If `False`, first look
+        for CB shapefiles
+        (https://www.census.gov/geographies/mapping-files/time-series/geo/carto-boundary-file.html),
+        which are more suitable for plotting maps, then fall back on the full
+        TIGER files
+        (https://www.census.gov/geographies/mapping-files/time-series/geo/tiger-line-file.html)
+        only if CB is not available. This is mainly set to `True` only
+        when `with_geometry_columns` is also set to `True`. The reason
+        is that the additional columns in the shapefiles are different
+        in the CB files than in the TIGER files.
     remove_water
         If `True` and if with_geometry=True, will query TIGER for AREAWATER shapefiles and
         remove water areas from returned geometry.
@@ -427,6 +719,25 @@ def download(
     -------
         A :py:class:`~pd.DataFrame` or `~gpd.GeoDataFrame` containing the requested US Census data.
     """
+    if dataset.startswith("lodes/"):
+        # Special case for the LODES data sets, which go down a completely
+        # different path.
+        if download_contained_within is not None:
+            raise ValueError(
+                "`download_contained_within` not supported for LODES data sets."
+            )
+
+        return download_lodes(
+            dataset,
+            vintage,
+            download_variables,
+            with_geometry=with_geometry,
+            with_geometry_columns=with_geometry_columns,
+            tiger_shapefiles_only=tiger_shapefiles_only,
+            remove_water=remove_water,
+            **kwargs,
+        )
+
     if download_contained_within is not None:
         # Put the contained_within context around it.
         return contained_within(
@@ -439,7 +750,10 @@ def download(
             leaves_of_group=leaves_of_group,
             set_to_nan=set_to_nan,
             skip_annotations=skip_annotations,
+            query_filter=query_filter,
             with_geometry=with_geometry,
+            with_geometry_columns=with_geometry_columns,
+            tiger_shapefiles_only=tiger_shapefiles_only,
             remove_water=remove_water,
             api_key=api_key,
             variable_cache=variable_cache,
@@ -490,9 +804,12 @@ def download(
             dataset,
             vintage,
             download_variables,
-            key=api_key,
+            api_key=api_key,
             census_variables=variable_cache,
+            query_filter=query_filter,
             with_geometry=with_geometry,
+            with_geometry_columns=with_geometry_columns,
+            tiger_shapefiles_only=tiger_shapefiles_only,
             row_keys=row_keys,
             **kwargs,
         )
@@ -513,7 +830,10 @@ def download(
         vintage,
         download_variables=download_variables,
         set_to_nan=set_to_nan,
+        query_filter=query_filter,
         with_geometry=with_geometry,
+        with_geometry_columns=with_geometry_columns,
+        tiger_shapefiles_only=tiger_shapefiles_only,
         remove_water=remove_water,
         api_key=api_key,
         variable_cache=variable_cache,
@@ -527,7 +847,10 @@ def _download_remote(
     *,
     download_variables: List[str],
     set_to_nan: Union[bool, Iterable[float]] = True,
+    query_filter: Optional[Dict[str, str]] = None,
     with_geometry: bool,
+    with_geometry_columns: bool,
+    tiger_shapefiles_only: bool,
     remove_water: bool,
     api_key: Optional[str],
     variable_cache: "VariableCache",
@@ -555,12 +878,35 @@ def _download_remote(
         values that the U.S. Census API sometimes returns. If `True`, then all
         values in :py:ref:`censusdis.values.ALL_SPECIAL_VALUES` will be replaced.
         If `False`, no replacements will be made.
+    query_filter
+        A dictionary of values to filter on. For example, if
+        `query_filter={'NAICS2017': '72251'}` then only rows
+        where the variable `NAICS2017` has a value of `'72251'`
+        will be returned.
+
+        This filtering is done on the server side, not the client
+        side, so it is far more efficient than querying without a
+        query filter and then manually filtering the results.
     with_geometry
         If `True` a :py:class:`gpd.GeoDataFrame` will be returned and each row
         will have a geometry that is a cartographic boundary suitable for platting
         a map. See https://www.census.gov/geographies/mapping-files/time-series/geo/cartographic-boundary.2020.html
         for details of the shapefiles that will be downloaded on your behalf to
         generate these boundaries.
+    with_geometry_columns
+        If `True` keep all the additional columns that come with shapefiles
+        downloaded to get geometry information.
+    tiger_shapefiles_only
+        If `True` only look for TIGER shapefiles. If `False`, first look
+        for CB shapefiles
+        (https://www.census.gov/geographies/mapping-files/time-series/geo/carto-boundary-file.html),
+        which are more suitable for plotting maps, then fall back on the full
+        TIGER files
+        (https://www.census.gov/geographies/mapping-files/time-series/geo/tiger-line-file.html)
+        only if CB is not available. This is mainly set to `True` only
+        when `with_geometry_columns` is also set to `True`. The reason
+        is that the additional columns in the shapefiles are different
+        in the CB files than in the TIGER files.
     api_key
         An optional API key. If you don't have or don't use a key, the number
         of calls you can make will be limited.
@@ -575,7 +921,12 @@ def _download_remote(
         either a `pd.DataFrame` or `gpd.GeoDataFrame`.
     """
     url, params, bound_path = census_table_url(
-        dataset, vintage, download_variables, api_key=api_key, **kwargs
+        dataset,
+        vintage,
+        download_variables,
+        query_filter=query_filter,
+        api_key=api_key,
+        **kwargs,
     )
     df_data = data_from_url(url, params)
 
@@ -603,7 +954,14 @@ def _download_remote(
         geo_level = bound_path.path_spec.path[-1]
         shapefile_scope = bound_path.bindings[bound_path.path_spec.path[0]]
 
-        gdf_data = add_geography(df_data, vintage, shapefile_scope, geo_level)
+        gdf_data = add_geography(
+            df_data,
+            vintage,
+            shapefile_scope,
+            geo_level,
+            with_geometry_columns=with_geometry_columns,
+            tiger_shapefiles_only=tiger_shapefiles_only,
+        )
 
         if remove_water:
             gdf_data = clip_water(gdf_data, vintage)
@@ -815,6 +1173,7 @@ def census_table_url(
     vintage: VintageType,
     download_variables: Iterable[str],
     *,
+    query_filter: Optional[Dict[str, str]] = None,
     api_key: Optional[str] = None,
     **kwargs: cgeo.InSpecType,
 ) -> Tuple[str, Mapping[str, str], cgeo.BoundGeographyPath]:
@@ -832,6 +1191,15 @@ def census_table_url(
         a timeseries data set, pass the string `'timeseries'`.
     download_variables
         The census variables to download, for example `["NAME", "B01001_001E"]`.
+    query_filter
+        A dictionary of values to filter on. For example, if
+        `query_filter={'NAICS2017': '72251'}` then only rows
+        where the variable `NAICS2017` has a value of `'72251'`
+        will be returned.
+
+        This filtering is done on the server side, not the client
+        side, so it is far more efficient than querying without a
+        query filter and then manually filtering the results.
     api_key
         An optional API key. If you don't have or don't use a key, the number
         of calls you can make will be limited.
@@ -849,7 +1217,7 @@ def census_table_url(
         dataset, vintage, list(download_variables), bound_path, api_key=api_key
     )
 
-    url, params = query_spec.table_url()
+    url, params = query_spec.table_url(query_filter=query_filter)
 
     return url, params, bound_path
 
@@ -862,14 +1230,47 @@ def _bind_path_if_possible(dataset, vintage, **kwargs):
     """
     bound_path = cgeo.PathSpec.partial_prefix_match(dataset, vintage, **kwargs)
     if bound_path is None:
-        raise CensusApiException(
-            f"Unable to match the geography specification {kwargs}.\n"
-            f"Supported geographies for dataset='{dataset}' in year={vintage} are:\n"
-            + "\n".join(
-                f"{path_spec}"
-                for path_spec in cgeo.geo_path_snake_specs(dataset, vintage).values()
+        if kwargs:
+            path_specs = cgeo.geo_path_snake_specs(dataset, vintage)
+            possible_path_spec_keys = set(
+                geo for path_name in path_specs.values() for geo in path_name
             )
-        )
+
+            non_geo_kwargs = [
+                k for k in kwargs.keys() if k not in possible_path_spec_keys
+            ]
+
+            if non_geo_kwargs:
+                msg = f"""
+The following arguments are not recognized as non-geographic arguments or goegraphic arguments
+for the dataset {dataset} in vintage {vintage}: '{"', '".join(non_geo_kwargs)}'.
+
+There are two reasons why this might happen:
+
+1. The arg(s) mentioned above are mispelled versions of named or geopgrahic arguments.
+2. The arg(s) mentioned above are valid geographic arguments for some data sets and
+   vintages, but not for {dataset} in vintage {vintage}.
+
+"""
+            else:
+                msg = f"""
+Unable to match the geography specification {kwargs}.
+
+"""
+
+            raise CensusApiException(
+                f"{msg}"
+                f"Supported geographies for dataset='{dataset}' in year={vintage} are:\n"
+                + "\n".join(
+                    f"{path_spec}"
+                    for path_spec in cgeo.geo_path_snake_specs(
+                        dataset, vintage
+                    ).values()
+                )
+            )
+        else:
+            bound_path = cgeo.BoundGeographyPath("000", cgeo.PathSpec.empty_path_spec())
+
     return bound_path
 
 
@@ -1084,7 +1485,10 @@ class ContainedWithin:
         leaves_of_group: Optional[Union[str, Iterable[str]]] = None,
         set_to_nan: Union[bool, Iterable[int]] = True,
         skip_annotations: bool = True,
+        query_filter: Optional[Dict[str, str]] = None,
         with_geometry: bool = False,
+        with_geometry_columns: bool = False,
+        tiger_shapefiles_only: bool = False,
         remove_water: bool = False,
         api_key: Optional[str] = None,
         variable_cache: Optional["VariableCache"] = None,
@@ -1127,12 +1531,35 @@ class ContainedWithin:
             annotations rather than actual values. See :py:meth:`VariableCache.group_variables`
             for more details. Variable names passed in `download_variables` are not
             affected by this flag.
+        query_filter
+            A dictionary of values to filter on. For example, if
+            `query_filter={'NAICS2017': '72251'}` then only rows
+            where the variable `NAICS2017` has a value of `'72251'`
+            will be returned.
+
+            This filtering is done on the server side, not the client
+            side, so it is far more efficient than querying without a
+            query filter and then manually filtering the results.
         with_geometry
             If `True` a :py:class:`gpd.GeoDataFrame` will be returned and each row
             will have a geometry that is a cartographic boundary suitable for platting
             a map. See https://www.census.gov/geographies/mapping-files/time-series/geo/cartographic-boundary.2020.html
             for details of the shapefiles that will be downloaded on your behalf to
             generate these boundaries.
+        with_geometry_columns
+            If `True` keep all the additional columns that come with shapefiles
+            downloaded to get geometry information.
+        tiger_shapefiles_only
+            If `True` only look for TIGER shapefiles. If `False`, first look
+            for CB shapefiles
+            (https://www.census.gov/geographies/mapping-files/time-series/geo/carto-boundary-file.html),
+            which are more suitable for plotting maps, then fall back on the full
+            TIGER files
+            (https://www.census.gov/geographies/mapping-files/time-series/geo/tiger-line-file.html)
+            only if CB is not available. This is mainly set to `True` only
+            when `with_geometry_columns` is also set to `True`. The reason
+            is that the additional columns in the shapefiles are different
+            in the CB files than in the TIGER files.
         remove_water
             If `True` and if with_geometry=True, will query TIGER for AREAWATER shapefiles and
             remove water areas from returned geometry.
@@ -1166,7 +1593,10 @@ class ContainedWithin:
             leaves_of_group=leaves_of_group,
             set_to_nan=set_to_nan,
             skip_annotations=skip_annotations,
+            query_filter=query_filter,
             with_geometry=True,
+            with_geometry_columns=with_geometry_columns,
+            tiger_shapefiles_only=tiger_shapefiles_only,
             remove_water=remove_water,
             api_key=api_key,
             variable_cache=variable_cache,
@@ -1250,7 +1680,11 @@ def contained_within(
 
 
 def add_inferred_geography(
-    df_data: pd.DataFrame, year: Optional[int] = None
+    df_data: pd.DataFrame,
+    year: Optional[int] = None,
+    *,
+    with_geometry_columns: bool = False,
+    tiger_shapefiles_only: bool = False,
 ) -> gpd.GeoDataFrame:
     """
     Infer the geography level of the given dataframe.
@@ -1271,6 +1705,20 @@ def add_inferred_geography(
         because they change over time. If `None`, look for a
         `'YEAR'` column in `df_data` and possibly add different
         geometries for different years as needed.
+    with_geometry_columns
+        If `True` keep all the additional columns that come with shapefiles
+        downloaded to get geometry information.
+    tiger_shapefiles_only
+        If `True` only look for TIGER shapefiles. If `False`, first look
+        for CB shapefiles
+        (https://www.census.gov/geographies/mapping-files/time-series/geo/carto-boundary-file.html),
+        which are more suitable for plotting maps, then fall back on the full
+        TIGER files
+        (https://www.census.gov/geographies/mapping-files/time-series/geo/tiger-line-file.html)
+        only if CB is not available. This is mainly set to `True` only
+        when `with_geometry_columns` is also set to `True`. The reason
+        is that the additional columns in the shapefiles are different
+        in the CB files than in the TIGER files.
 
     Returns
     -------
@@ -1285,9 +1733,18 @@ def add_inferred_geography(
             )
 
         return gpd.GeoDataFrame(
-            df_data.groupby("YEAR", group_keys=False)
-            .apply(lambda df_group: add_inferred_geography(df_group, df_group.name))
-            .reset_index(drop=True)
+            df_data.groupby("YEAR", group_keys=True, sort=False)
+            .apply(
+                lambda df_group: add_inferred_geography(
+                    df_group,
+                    df_group.name,
+                    with_geometry_columns=with_geometry_columns,
+                    tiger_shapefiles_only=tiger_shapefiles_only,
+                ),
+                include_groups=False,
+            )
+            .reset_index(level=1, drop=True)
+            .reset_index(drop=False)
         )
 
     geo_level = infer_geo_level(year, df_data)
@@ -1301,7 +1758,14 @@ def add_inferred_geography(
 
     if shapefile_scope is not None:
         # The scope is the same across the board.
-        gdf = add_geography(df_data, year, shapefile_scope, geo_level)
+        gdf = add_geography(
+            df_data,
+            year,
+            shapefile_scope,
+            geo_level,
+            with_geometry_columns=with_geometry_columns,
+            tiger_shapefiles_only=tiger_shapefiles_only,
+        )
         return gdf
 
     # We have to group by different values of the shapefile
@@ -1310,9 +1774,20 @@ def add_inferred_geography(
     shapefile_scope_column = shapefile_scope_columns[0]
 
     df_with_geo = (
-        df_data.groupby(shapefile_scope_column, group_keys=False)
-        .apply(lambda g: add_geography(g, year, g.name, geo_level))
-        .reset_index(drop=True)
+        df_data.set_index(shapefile_scope_column)
+        .groupby(level=shapefile_scope_column, group_keys=True, sort=False)
+        .apply(
+            lambda g: add_geography(
+                g,
+                year,
+                g.name,
+                geo_level,
+                with_geometry_columns=with_geometry_columns,
+                tiger_shapefiles_only=tiger_shapefiles_only,
+            )
+        )
+        .reset_index(level=1, drop=True)
+        .reset_index(drop=False)
     )
 
     gdf = gpd.GeoDataFrame(df_with_geo)
